/*************************************************************************
*
* DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
*
* Copyright (c) 2011 Linagora
* Maxime de Roucy <maxime.deroucy@linagora.com>
* licence : GPLv3
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
************************************************************************/

// inspired from http://www.blizzhackers.cc/viewtopic.php?p=2483118
// by Darawk <darawk@gmail.com>

#include <windows.h>
#include <tlhelp32.h>
#include <shlwapi.h>
#include <psapi.h>

#include <fstream>
#include <string>
#include <vector>

#define ERROR_TITLE "dll_loader error"
#define PROCESS_NAME "soffice.bin"
#define DLL_LIST_LIBRE_NAME "dll_loader_libreoffice.txt"
#define DLL_LIST_OPEN_NAME "dll_loader_openoffice.txt"
#define WAIT_TIME 10

// minimum priveleges to load our dlls
#define CREATE_THREAD_ACCESS (PROCESS_CREATE_THREAD | PROCESS_QUERY_INFORMATION | PROCESS_VM_OPERATION | PROCESS_VM_WRITE | PROCESS_VM_READ)

using namespace std;

bool IsWindowsNT();
void GetLibreOpenOfficeProcessId(vector<unsigned long>& pids);
HANDLE OpenProcess(unsigned long pid);
void LoadDllInProcess(HANDLE process, LPTHREAD_START_ROUTINE LoadLibraryA, string dll_list_file_name);

// main function
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	if(!IsWindowsNT())
	{
		MessageBox(NULL, "Your system isn't supported", ERROR_TITLE, MB_OK | MB_ICONERROR );
		return 1;
	}

	// search and open the process
	vector<unsigned long> pids;
	pids.reserve(2);

	for(unsigned int i = 0 ; i < WAIT_TIME ; ++i)
	{
		GetLibreOpenOfficeProcessId(pids);

		if (pids.size() == 2)
			break;
		else
			Sleep(1000);
	}
	
	if (pids.empty())
	{
		MessageBox(NULL, "Unable to find LibreOffice nor OpenOffice Quickstarter", ERROR_TITLE, MB_OK | MB_ICONERROR );
		exit(1);
	}

	// open both process
	HANDLE Proc_1 = OpenProcess(pids[0]);
	HANDLE Proc_2 = (pids.size() == 2 ? OpenProcess(pids[1]) : NULL);

	// name the process (which one is libreoffice ?)
	char* full_path_exe_name = new char[255];
	unsigned int full_path_exe_name_size = GetModuleFileNameEx(Proc_1, NULL, full_path_exe_name, 255);

	if (full_path_exe_name_size == 0 || full_path_exe_name_size == 255)
	{
		string error_message = "Unable to find the full path of the exe file of process PID : ";
		error_message += pids[0];
		MessageBox(NULL, error_message.c_str(), ERROR_TITLE, MB_OK | MB_ICONERROR );
		exit(1);
	}

	string proc_1_dll_list, proc_2_dll_list;
	if (StrStrI(full_path_exe_name, "LibreOffice"))
	{
		proc_1_dll_list = DLL_LIST_LIBRE_NAME;
		proc_2_dll_list = DLL_LIST_OPEN_NAME;
	} else {
		proc_1_dll_list = DLL_LIST_OPEN_NAME;
		proc_2_dll_list = DLL_LIST_LIBRE_NAME;
	}
	
	// locate the LoadLibraryA (ANSI) function in the kernel32 dll
	LPTHREAD_START_ROUTINE LoadLibraryA = reinterpret_cast<LPTHREAD_START_ROUTINE>(GetProcAddress(GetModuleHandle("kernel32.dll"), "LoadLibraryA"));
	
	LoadDllInProcess(Proc_1, LoadLibraryA, proc_1_dll_list);
	CloseHandle(Proc_1);

	if (Proc_2)
	{
		LoadDllInProcess(Proc_2, LoadLibraryA, proc_2_dll_list);
		CloseHandle(Proc_2);
	}

	return 0;
}

// function verifying that Windows is an NT-based system.
bool IsWindowsNT()
{
	return (GetVersion() < 0x80000000);
}

// function to obtaining a process ID of Libreoffice and OpenOffice
void GetLibreOpenOfficeProcessId(vector<unsigned long>& pids)
{
	PROCESSENTRY32 pe;

	// Takes a snapshot of all the processes in the systeme
	HANDLE thSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);

	if(thSnapshot == INVALID_HANDLE_VALUE)
	{
		MessageBox(NULL, "Unable to obtain the process list.", ERROR_TITLE, MB_OK | MB_ICONERROR );
		exit(1);
	}

	// needed by Process32First
	pe.dwSize = sizeof(PROCESSENTRY32);

	bool retval = Process32First(thSnapshot, &pe);

	while(retval)
	{
		if(StrStrI(pe.szExeFile, PROCESS_NAME))
		{
			if (pids.empty())
			{
				pids.push_back(pe.th32ProcessID);
			}
			else if (pids.size() == 1 && pids[0] != pe.th32ProcessID)
			{
				pids.push_back(pe.th32ProcessID);
				return;
			}
		}

		retval = Process32Next(thSnapshot,&pe);
	}
}

HANDLE OpenProcess(unsigned long pid)
{
	HANDLE Proc = OpenProcess(CREATE_THREAD_ACCESS, FALSE, pid);

	if(!Proc)
	{
		string error_message = "Unable to open the process PID : ";
		error_message += pid;
		MessageBox(NULL, error_message.c_str(), ERROR_TITLE, MB_OK | MB_ICONERROR );
		exit(1);
	}
	
	return Proc;
}

void LoadDllInProcess(HANDLE process, LPTHREAD_START_ROUTINE LoadLibraryA, string dll_list_file_name)
{
	string error_message;
	ifstream dll_list_file;
	dll_list_file.open(dll_list_file_name.c_str());

	if (dll_list_file.fail())
	{
		error_message = "Unable to open the ";
		error_message += dll_list_file_name;
		error_message += " file name";
		MessageBox(NULL, error_message.c_str(), ERROR_TITLE, MB_OK | MB_ICONERROR );
		return;
	}

	string current_dll_name("");
	getline(dll_list_file, current_dll_name);

	while (!dll_list_file.fail())
	{
		// allocate space in the remote process to put the dllName string
		LPVOID RemoteString = static_cast<LPVOID>(VirtualAllocEx(process, NULL, current_dll_name.length(), MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE));

		if (!RemoteString)
			MessageBox(NULL, "Unable to allocate space in the remote process to put the dllName string.", ERROR_TITLE, MB_OK | MB_ICONERROR );

		// Write the dllName string in the memory space allocated above
		if (!WriteProcessMemory(process, RemoteString, current_dll_name.c_str(), current_dll_name.length(), NULL))
		{
			error_message = "Unable to write the dllName ";
			error_message += current_dll_name;
			error_message += " in the memory space of the process.";
			MessageBox(NULL, error_message.c_str(), ERROR_TITLE, MB_OK | MB_ICONERROR );

			getline(dll_list_file, current_dll_name);
			continue;
		}

		// create the remote thread that will launch the LoadLibraryA function that will load the dll.
		if (!CreateRemoteThread(process, NULL, 0, LoadLibraryA, RemoteString, 0, NULL))
		{
			error_message = "Unable create the remote thread for :";
			error_message += current_dll_name;
			MessageBox(NULL, error_message.c_str(), ERROR_TITLE, MB_OK | MB_ICONERROR );
		}

		getline(dll_list_file, current_dll_name);
	}

	dll_list_file.close();
}
